#!/usr/bin/python3
"""This file provides logging utilities"""

import datetime
import logging
import re
import sys
from typing import Optional

import maven_check_versions.config as _config
import requests
from maven_check_versions.config import Config, Arguments


class Formatter(logging.Formatter):
    """
    Formatter with microseconds.
    """

    def formatTime(self, record, datefmt=None):  # pragma: no cover # noqa: N802
        """
        Formats the timestamp of a log record with microsecond precision.

        Args:
            record: The log record to format.
            datefmt (str, optional): Date format string (unused, defaults to None).

        Returns:
            str: The formatted timestamp as a string.
        """
        value: datetime.datetime = datetime.datetime.fromtimestamp(record.created)
        return value.strftime('%Y-%m-%d %H:%M:%S.%f')


def configure_logging(arguments: Arguments) -> None:
    """
    Configures the logging system to output to stdout and optionally to a file.
    Sets the log level to INFO and applies a custom formatter with timestamps.

    Args:
        arguments (Arguments): Command-line arguments, which may include 'logfile_off'
                            to disable file logging and 'log_file' to specify the log file path.
    """
    log_format = '%(asctime)s %(levelname)s: %(message)s'
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.formatter = Formatter(fmt=log_format)
    handlers: list = [stream_handler]

    if not arguments.get('logfile_off'):
        if (log_file_path := arguments.get('log_file')) is None:
            log_file_path = 'maven_check_versions.log'
        file_handler = logging.FileHandler(log_file_path)
        file_handler.formatter = Formatter(fmt=log_format)
        handlers.append(file_handler)

    logging.basicConfig(  # NOSONAR
        level=(arguments.get('log_level') or 'info').upper(), handlers=handlers
    )


def log_skip_if_required(
        config: Config, arguments: Arguments, group: str, artifact: str, version: Optional[str]
) -> None:
    """
    Logs a skipped dependency if required.

    Args:
        config (Config): Parsed YAML as dict.
        arguments (Arguments): Command-line arguments.
        group (str): Group ID.
        artifact (str): Artifact ID.
        version (Optional[str]): Dependency version.
    """
    if _config.get_config_value(config, arguments, 'show_skip'):
        logging.warning(f"Skip: {group}:{artifact}:{version}")


def log_search_if_required(
        config: Config, arguments: Arguments, group: str, artifact: str, version: Optional[str]
) -> None:
    """
    Logs a dependency search action if required.

    Args:
        config (Config): Parsed YAML as dict.
        arguments (Arguments): Command-line arguments.
        group (str): Group ID.
        artifact (str): Artifact ID.
        version (Optional[str]): Dependency version (Maybe None or a placeholder).
    """
    if _config.get_config_value(config, arguments, 'show_search'):
        if version is None or re.match(r'^\${([^}]+)}$', version):
            logging.warning(f"Search: {group}:{artifact}:{version}")
        else:
            logging.info(f"Search: {group}:{artifact}:{version}")


def log_invalid_if_required(
        config: Config, arguments: Arguments, response: requests.Response, group: str,
        artifact: str, item: str, invalid_flag: bool
) -> None:
    """
    Logs invalid versions if required.

    Args:
        config (Config): Parsed YAML as dict.
        arguments (Arguments): Command-line arguments.
        response (requests.Response): Repository response.
        group (str): Group ID.
        artifact (str): Artifact ID.
        item (str): Version being checked.
        invalid_flag (bool): Flag indicating invalid versions have been logged.
    """
    if _config.get_config_value(config, arguments, 'show_invalid'):
        if not invalid_flag:
            logging.info(response.url)
        logging.warning(f"Invalid: {group}:{artifact}:{item}")



#!/usr/bin/python3
"""This file provides cache utilities"""
import json
import logging
import os
import threading
import time
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import asdict, is_dataclass
from pathlib import Path
from typing import Optional, Dict, Any

import maven_check_versions.config as _config
import pymemcache
import redis
import tarantool
from maven_check_versions.config import Config, Arguments

_ARTIFACTS_KEY = 'cache_maven_check_versions_artifacts'
_VULNERABILITIES_KEY = 'cache_maven_check_versions_vulnerabilities'
_HOST = 'localhost'
_REDIS_PORT = 6379
_TARANTOOL_PORT = 3301
_MEMCACHED_PORT = 11211

update_cache_artifact_lock = threading.Lock()


class DCJSONEncoder(json.JSONEncoder):  # pragma: no cover
    """
    Custom JSON encoder for serializing dataclasses.
    """

    def default(self, obj):
        """
        Encodes dataclass objects to JSON by converting them to dictionaries.

        Args:
            obj: The object to encode.

        Returns:
            dict: The encoded object as a dictionary, or delegates to the parent encoder.
        """
        try:
            return asdict(obj) if is_dataclass(obj) else super().default(obj)
        except Exception as e:
            logging.error(f"Failed to encode dataclass: {e}")
            return None


class _CacheBackend(ABC):
    """
    Abstract base class for cache backend implementations.
    """

    @abstractmethod
    def load(self, config: Config, arguments: Arguments, section: str) -> Dict[str, Any]:  # pragma: no cover
        """
        Loads the cache data from the specified backend.

        Args:
            config (Config): Configuration dictionary parsed from YAML.
            arguments (Arguments): Command-line arguments.
            section (str): Configuration section to use (e.g., 'base' or 'vulnerability').

        Returns:
            Dict[str, Any]: The cache data dictionary. Returns an empty dictionary if loading
                fails or no data is available.
        """
        pass

    @abstractmethod
    def save(  # pragma: no cover
            self, config: Config, arguments: Arguments, cache_data: Dict[str, Any], section: str
    ) -> None:
        """
        Saves the cache data.

        Args:
            config (Config): Configuration dictionary parsed from YAML.
            arguments (Arguments): Command-line arguments.
            cache_data (Dict[str, Any]): Cache data to save.
            section (str): Configuration section to use (e.g., 'base' or 'vulnerability').
        """
        pass


class _CacheBackendRegistry:
    """
    Registry for managing cache backend implementations.
    """
    _backends: dict[str, _CacheBackend] = {}

    @classmethod
    def register(cls, name: str, backend: _CacheBackend):
        """
        Register a cache backend with a given name.
        """
        cls._backends[name] = backend

    @classmethod
    def get(cls, name: str) -> _CacheBackend:
        """
        Retrieve a cache backend by name, defaulting to JSON if not found.
        """
        return cls._backends.get(name, _JSONCacheBackend())


class _JSONCacheBackend(_CacheBackend):
    """
    Backend for caching data in json files.
    """

    def load(self, config: Config, arguments: Arguments, section: str) -> Dict[str, Any]:
        cache_file = _config.get_config_value(
            config, arguments, 'cache_file', section=section,
            default=(_VULNERABILITIES_KEY if section == 'vulnerability' else _ARTIFACTS_KEY) + '.json')

        if os.path.exists(cache_file):
            try:
                logging.info(f"Load Cache file: {Path(cache_file).absolute()}")
                with open(cache_file, encoding='utf-8') as cf:
                    return json.load(cf)
            except json.JSONDecodeError as e:
                logging.error(f"Failed to decode JSON cache data: {e}")
        return {}

    def save(
            self, config: Config, arguments: Arguments, cache_data: Dict[str, Any], section: str
    ) -> None:
        cache_file = _config.get_config_value(
            config, arguments, 'cache_file', section=section,
            default=(_VULNERABILITIES_KEY if section == 'vulnerability' else _ARTIFACTS_KEY) + '.json')

        try:
            logging.info(f"Save Cache file: {Path(cache_file).absolute()}")
            with open(cache_file, 'w', encoding='utf-8') as cf:
                cf.write(json.dumps(cache_data, cls=DCJSONEncoder, indent=2))
        except Exception as e:
            logging.error(f"Failed to save cache to JSON file {cache_file}: {e}")


class _RedisCacheBackend(_CacheBackend):
    """
    Backend for caching data in Redis.
    """

    @staticmethod
    def _config(config: Config, arguments: Arguments, section: str) -> tuple:
        """
        Retrieves the Redis connection parameters from the configuration.

        Args:
            config (Config): Configuration dictionary parsed from YAML.
            arguments (Arguments): Command-line arguments.
            section (str): Configuration section to use (e.g., 'base' or 'vulnerability').

        Returns:
            tuple: A tuple containing (host, port, key, user, password) for Redis connection.
        """
        return (
            _config.get_config_value(
                config, arguments, 'redis_host', section=section, default=_HOST),
            _config.get_config_value(
                config, arguments, 'redis_port', section=section, default=_REDIS_PORT),
            _config.get_config_value(
                config, arguments, 'redis_key', section=section,
                default=_VULNERABILITIES_KEY if section == 'vulnerability' else _ARTIFACTS_KEY),
            _config.get_config_value(config, arguments, 'redis_user', section=section),
            _config.get_config_value(config, arguments, 'redis_password', section=section)
        )

    @contextmanager
    def _connection(
            self, host: str, port: int, user: Optional[str], password: Optional[str]
    ):
        """
        Context manager for Redis connection, ensuring proper cleanup.

        Args:
            host (str): Redis server host.
            port (int): Redis server port.
            user (Optional[str]): Redis username, if required.
            password (Optional[str]): Redis password, if required.

        Yields:
            redis.Redis: An instance of the Redis client.

        Raises:
            redis.ConnectionError: If the connection to Redis fails.
            redis.RedisError: If an error occurs during Redis operations.
        """
        inst = redis.Redis(host=host, port=port, username=user, password=password, decode_responses=True)
        try:
            yield inst
        finally:
            inst.close()

    def load(self, config: Config, arguments: Arguments, section: str) -> Dict[str, Any]:
        try:
            host, port, ckey, user, password = self._config(config, arguments, section)

            with self._connection(host, port, user, password) as inst:
                cache_data: Dict[str, Any] = {}
                if data := inst.hgetall(ckey):
                    for key, value in data.items():
                        try:
                            cache_data[key] = json.loads(value)
                        except json.JSONDecodeError as e:
                            logging.error(f"Failed to decode Redis data for key {key}: {e}")
                return cache_data

        except redis.ConnectionError as e:  # pragma: no cover
            logging.error(f"Redis connection failed: {e}")
        except redis.RedisError as e:  # pragma: no cover
            logging.error(f"Redis error: {e}")
        except Exception as e:
            logging.error(f"Failed to load cache from Redis: {e}")
        return {}

    def save(
            self, config: Config, arguments: Arguments, cache_data: Dict[str, Any], section: str
    ) -> None:
        try:
            host, port, ckey, user, password = self._config(config, arguments, section)

            with self._connection(host, port, user, password) as inst:
                for key, value in cache_data.items():
                    try:
                        inst.hset(ckey, key, json.dumps(value, cls=DCJSONEncoder))
                    except redis.RedisError as e:  # pragma: no cover
                        logging.error(f"Failed to save cache to Redis for key {key}: {e}")

        except redis.ConnectionError as e:  # pragma: no cover
            logging.error(f"Redis connection failed: {e}")
        except redis.RedisError as e:  # pragma: no cover
            logging.error(f"Redis error: {e}")
        except Exception as e:
            logging.error(f"Failed to save cache to Redis: {e}")


class _TarantoolCacheBackend(_CacheBackend):
    """
    Backend for caching data in Tarantool.
    """

    @staticmethod
    def _config(config: Config, arguments: Arguments, section: str) -> tuple:
        """
        Retrieves the Tarantool connection parameters from the configuration.

        Args:
            config (Config): Configuration dictionary parsed from YAML.
            arguments (Arguments): Command-line arguments.
            section (str): Configuration section to use (e.g., 'base' or 'vulnerability').

        Returns:
            tuple: A tuple containing (host, port, space, user, password) for Tarantool connection.
        """
        return (
            _config.get_config_value(
                config, arguments, 'tarantool_host', section=section, default=_HOST),
            _config.get_config_value(
                config, arguments, 'tarantool_port', section=section, default=_TARANTOOL_PORT),
            _config.get_config_value(
                config, arguments, 'tarantool_space', section=section,
                default=_VULNERABILITIES_KEY if section == 'vulnerability' else _ARTIFACTS_KEY),
            _config.get_config_value(config, arguments, 'tarantool_user', section=section),
            _config.get_config_value(config, arguments, 'tarantool_password', section=section)
        )

    @contextmanager
    def _connection(
            self, host: str, port: int, user: Optional[str], password: Optional[str]
    ):
        """
        Context manager for Tarantool connection, ensuring proper cleanup.

        Args:
            host (str): Tarantool server host.
            port (int): Tarantool server port.
            user (Optional[str]): Tarantool username, if required.
            password (Optional[str]): Tarantool password, if required.

        Yields:
            tarantool.Connection: An instance of the Tarantool connection.

        Raises:
            tarantool.DatabaseError: If an error occurs during Tarantool operations.
        """
        conn = tarantool.Connection(host, port, user=user, password=password)
        try:
            yield conn
        finally:
            conn.close()

    def load(self, config: Config, arguments: Arguments, section: str) -> Dict[str, Any]:
        try:
            host, port, space, user, password = self._config(config, arguments, section)

            with self._connection(host, port, user, password) as conn:
                cache_data: Dict[str, Any] = {}
                if data := conn.select(space):
                    for item in data:
                        try:
                            cache_data[item[0]] = json.loads(item[1])
                        except json.JSONDecodeError as e:
                            logging.error(f"Failed to decode Tarantool data for key {item[0]}: {e}")
                return cache_data

        except tarantool.DatabaseError as e:  # pragma: no cover
            logging.error(f"Tarantool error: {e}")
        except Exception as e:
            logging.error(f"Failed to load cache from Tarantool: {e}")
        return {}

    def save(
            self, config: Config, arguments: Arguments, cache_data: Dict[str, Any], section: str
    ) -> None:
        try:
            host, port, space, user, password = self._config(config, arguments, section)

            with self._connection(host, port, user, password) as conn:
                space = conn.space(space)
                for key, value in cache_data.items():
                    space.replace((key, json.dumps(value, cls=DCJSONEncoder)))

        except tarantool.DatabaseError as e:  # pragma: no cover
            logging.error(f"Tarantool error: {e}")
        except Exception as e:
            logging.error(f"Failed to save cache to Tarantool: {e}")


class _MemcachedCacheBackend(_CacheBackend):
    """
    Backend for caching data in Memcached.
    """

    @staticmethod
    def _config(config: Config, arguments: Arguments, section: str) -> tuple:
        """
        Retrieves the Memcached connection parameters from the configuration.

        Args:
            config (Config): Configuration dictionary parsed from YAML.
            arguments (Arguments): Command-line arguments.
            section (str): Configuration section to use (e.g., 'base' or 'vulnerability').

        Returns:
            tuple: A tuple containing (host, port, key) for Memcached connection.
        """
        return (
            _config.get_config_value(
                config, arguments, 'memcached_host', section=section, default=_HOST),
            _config.get_config_value(
                config, arguments, 'memcached_port', section=section, default=_MEMCACHED_PORT),
            _config.get_config_value(
                config, arguments, 'memcached_key', section=section,
                default=_VULNERABILITIES_KEY if section == 'vulnerability' else _ARTIFACTS_KEY)
        )

    @contextmanager
    def _connection(self, host: str, port: int):
        """
        Context manager for Memcached connection, ensuring proper cleanup.

        Args:
            host (str): Memcached server host.
            port (int): Memcached server port.

        Yields:
            pymemcache.client.base.Client: An instance of the Memcached client.

        Raises:
            pymemcache.exceptions.MemcacheError: If an error occurs during Memcached operations.
        """
        client = pymemcache.client.base.Client((host, port))
        try:
            yield client
        finally:
            client.close()

    def load(self, config: Config, arguments: Arguments, section: str) -> Dict[str, Any]:
        try:
            host, port, key = self._config(config, arguments, section)

            with self._connection(host, port) as client:
                if data := client.get(key):
                    try:
                        return json.loads(data)
                    except json.JSONDecodeError as e:
                        logging.error(f"Failed to decode Memcached data: {e}")

        except pymemcache.exceptions.MemcacheError as e:  # pragma: no cover
            logging.error(f"Memcached error: {e}")
        except Exception as e:
            logging.error(f"Failed to load cache from Memcached: {e}")
        return {}

    def save(
            self, config: Config, arguments: Arguments, cache_data: Dict[str, Any], section: str
    ) -> None:
        try:
            host, port, key = self._config(config, arguments, section)

            with self._connection(host, port) as client:
                client.set(key, json.dumps(cache_data, cls=DCJSONEncoder))

        except pymemcache.exceptions.MemcacheError as e:  # pragma: no cover
            logging.error(f"Memcached error: {e}")
        except Exception as e:
            logging.error(f"Failed to save cache to Memcached: {e}")


_CacheBackendRegistry.register('json', _JSONCacheBackend())
_CacheBackendRegistry.register('redis', _RedisCacheBackend())
_CacheBackendRegistry.register('tarantool', _TarantoolCacheBackend())
_CacheBackendRegistry.register('memcached', _MemcachedCacheBackend())


def load_cache(config: Config, arguments: Arguments, section: str = 'base') -> Dict[str, Any]:
    """
    Loads the cache data from the specified backend based on the configuration.
    Supports JSON, Redis, Tarantool, and Memcached backends.

    Args:
        config (Config): Configuration dictionary parsed from YAML.
        arguments (Arguments): Command-line arguments.
        section (str, optional): Configuration section to use, such as 'base' or 'vulnerability'.
            Defaults to 'base'.

    Returns:
        Dict[str, Any]: Cache data as a dictionary.
            Returns an empty dictionary if the backend fails or no data is available.
            If the specified backend is not found, defaults to JSON backend.
    """
    key = _config.get_config_value(config, arguments, 'cache_backend', section=section)
    if backend := _CacheBackendRegistry.get(key):
        return backend.load(config, arguments, section)
    else:  # pragma: no cover
        raise AssertionError('Invalid cache backend')


def save_cache(
        config: Config, arguments: Arguments, cache_data: Optional[Dict[str, Any]],
        section: str = 'base'
) -> None:
    """
    Saves the cache data to the specified backend based on the configuration.
    Supports JSON, Redis, Tarantool, and Memcached backends.

    Args:
        config (Config): Configuration dictionary parsed from YAML.
        arguments (Arguments): Command-line arguments.
        cache_data (Optional[Dict[str, Any]]): Cache data to save.
        section (str, optional): Configuration section to use, such as 'base' or 'vulnerability'.
            Defaults to 'base'.
    """
    if cache_data is not None:
        key = _config.get_config_value(config, arguments, 'cache_backend', section=section)
        if backend := _CacheBackendRegistry.get(key):
            backend.save(config, arguments, cache_data, section)
        else:  # pragma: no cover
            raise AssertionError('Invalid cache backend')


def process_cache_artifact(
        config: Config, arguments: Arguments, cache_data: Optional[Dict[str, Any]],
        artifact: str, group: str, version: Optional[str]
) -> bool:
    """
    Checks if the cached data for the specified artifact is valid and up-to-date.

    Args:
        config (Config): Configuration dictionary parsed from YAML.
        arguments (Arguments): Command-line arguments.
        cache_data (Optional[Dict[str, Any]]): The cache data dictionary containing artifact information.
        artifact (str): The artifact ID of the dependency.
        group (str): The group ID of the dependency.
        version (Optional[str]): The current version of the artifact, or None if not specified or unresolved.

    Returns:
        bool: True if the cache exists and either the cached version matches the provided version
            or the cache timestamp is within the configured time threshold, False otherwise.
    """
    if cache_data is None or (data := cache_data.get(f"{group}:{artifact}")) is None:
        return False
    cached_time, cached_version, cached_key, cached_date, cached_versions = data
    if cached_version == version:
        return True

    ct_threshold = int(_config.get_config_value(config, arguments, 'cache_time', default=0))

    if ct_threshold == 0 or time.time() - cached_time < ct_threshold:
        message_format = '*{}: {}:{}:{}, last versions: {}, modified:{}.'
        logging.info(message_format.format(
            cached_key, group, artifact, version, ', '.join(cached_versions),
            cached_date if cached_date is not None else '').rstrip())
        return True
    return False


def update_cache_artifact(
        cache_data: Optional[Dict[str, Any]], versions: list, artifact: str, group,
        item: str, last_modified: Optional[str], section_key: str
) -> None:
    """
    Updates the cache dictionary with the latest data for the specified artifact.

    Args:
        cache_data (Optional[Dict[str, Any]]): The cache dictionary to update, or None if caching is disabled.
        versions (list): List of available versions for the artifact.
        artifact (str): The artifact ID of the dependency.
        group (str): The group ID of the dependency.
        item (str): The current version of the artifact being processed.
        last_modified (Optional[str]):
            The last modified date of the artifact in ISO format, or None if unavailable.
        section_key (str): The repository section key from the configuration.
    """
    if cache_data is not None:
        with update_cache_artifact_lock:
            value = (int(time.time()), item, section_key, last_modified, versions[:5])
            cache_data[f"{group}:{artifact}"] = value



#!/usr/bin/python3
"""This file provides utility functions"""

import logging
import os
import re
# noinspection PyPep8Naming
import xml.etree.ElementTree as ET
from argparse import ArgumentParser
from typing import Optional

import dateutil.parser as parser
import maven_check_versions.cache as _cache
import maven_check_versions.config as _config
import maven_check_versions.logutils as _logutils
import requests
from maven_check_versions.config import Config, Arguments


def parse_command_line() -> Arguments:
    """
    Parses the command-line arguments and returns them as an Arguments object.

    Returns:
        Arguments: An object containing the parsed command-line arguments,
                    wrapping a dictionary of argument key-value pairs.
    """
    argument_parser = ArgumentParser(prog='maven_check_versions')
    add_general_args(argument_parser)
    add_cache_args(argument_parser)
    add_logging_args(argument_parser)
    add_fail_mode_args(argument_parser)
    add_search_args(argument_parser)
    add_auth_args(argument_parser)
    add_threading_args(argument_parser)
    return Arguments(vars(argument_parser.parse_args()))


def add_general_args(argument_parser: ArgumentParser) -> None:
    """
    Adds general command-line arguments to the argument parser.
    These include options for CI mode, POM file path, artifact to find, and configuration file path.

    Args:
        argument_parser (ArgumentParser): The ArgumentParser instance to which general arguments are added.
    """
    argument_parser.add_argument('-ci', '--ci_mode', help='Enable CI Mode', action='store_true', default=False)
    argument_parser.add_argument('-pf', '--pom_file', help='Path to POM File')
    argument_parser.add_argument('-fa', '--find_artifact', help='Artifact to find')
    argument_parser.add_argument('-cfg', '--config_file', help='Path to Config File')
    argument_parser.add_argument('-ll', '--log_level', help='Logging level', default=None)


def add_cache_args(argument_parser: ArgumentParser) -> None:
    """
    Adds cache-related arguments to the parser.

    Args:
        argument_parser (ArgumentParser): The argument parser to which arguments are added.
    """
    argument_parser.add_argument('-co', '--cache_off', help='Disable Cache', action='store_true', default=None)
    argument_parser.add_argument('-cf', '--cache_file', help='Path to Cache File')
    argument_parser.add_argument('-ct', '--cache_time', help='Cache expiration time in seconds')
    argument_parser.add_argument('-cb', '--cache_backend', help='Cache backend')

    argument_parser.add_argument('-rsh', '--redis_host', help='Redis host', default=None)
    argument_parser.add_argument('-rsp', '--redis_port', help='Redis port', default=None)
    argument_parser.add_argument('-rsk', '--redis_key', help='Redis key', default=None)
    argument_parser.add_argument('-rsu', '--redis_user', help='Redis user', default=None)
    argument_parser.add_argument('-rsup', '--redis_password', help='Redis password', default=None)

    argument_parser.add_argument('-tlh', '--tarantool_host', help='Tarantool host', default=None)
    argument_parser.add_argument('-tlp', '--tarantool_port', help='Tarantool port', default=None)
    argument_parser.add_argument('-tls', '--tarantool_space', help='Tarantool space', default=None)
    argument_parser.add_argument('-tlu', '--tarantool_user', help='Tarantool user', default=None)
    argument_parser.add_argument('-tlup', '--tarantool_password', help='Tarantool password', default=None)

    argument_parser.add_argument('-mch', '--memcached_host', help='Memcached host', default=None)
    argument_parser.add_argument('-mcp', '--memcached_port', help='Memcached port', default=None)
    argument_parser.add_argument('-mck', '--memcached_key', help='Memcached key', default=None)


def add_logging_args(argument_parser: ArgumentParser) -> None:
    """
    Adds logging-related arguments to the parser.

    Args:
        argument_parser (ArgumentParser): The argument parser to which arguments are added.
    """
    argument_parser.add_argument('-lfo', '--logfile_off', help='Disable Log file', action='store_true', default=None)
    argument_parser.add_argument('-lf', '--log_file', help='Path to Log File')


def add_fail_mode_args(argument_parser: ArgumentParser) -> None:
    """
    Adds fail mode-related arguments to the parser.

    Args:
        argument_parser (ArgumentParser): The argument parser to which arguments are added.
    """
    argument_parser.add_argument('-fm', '--fail_mode', help='Enable Fail Mode', action='store_true', default=None)
    argument_parser.add_argument('-mjv', '--fail_major', help='Major version threshold for failure')
    argument_parser.add_argument('-mnv', '--fail_minor', help='Minor version threshold for failure')


def add_search_args(argument_parser: ArgumentParser) -> None:
    """
    Adds search-related arguments to the parser.

    Args:
        argument_parser (ArgumentParser): The argument parser to which arguments are added.
    """
    argument_parser.add_argument('-sp', '--search_plugins', help='Search plugins', action='store_true', default=None)
    argument_parser.add_argument('-sm', '--process_modules', help='Process modules', action='store_true', default=None)
    argument_parser.add_argument('-sk', '--show_skip', help='Show Skip', action='store_true', default=None)
    argument_parser.add_argument('-ss', '--show_search', help='Show Search', action='store_true', default=None)
    argument_parser.add_argument(
        '-ev', '--empty_version', help='Allow empty version', action='store_true', default=None)
    argument_parser.add_argument('-si', '--show_invalid', help='Show Invalid', action='store_true', default=None)


def add_auth_args(argument_parser: ArgumentParser) -> None:
    """
    Adds authentication-related arguments to the parser.

    Args:
        argument_parser (ArgumentParser): The argument parser to which arguments are added.
    """
    argument_parser.add_argument('-un', '--user', help='Basic Auth user')
    argument_parser.add_argument('-up', '--password', help='Basic Auth password')


def add_threading_args(argument_parser: ArgumentParser) -> None:
    """
    Adds threading-related arguments to the parser.

    Args:
        argument_parser (ArgumentParser): The argument parser to which arguments are added.
    """
    argument_parser.add_argument('-th', '--threading', help='Enable threading', action='store_true', default=None)
    argument_parser.add_argument('-mt', '--max_threads', help='Maximum number of threads', type=int)


def get_artifact_name(root: ET.Element, ns_mapping: dict) -> str:
    """
    Extracts the groupId and artifactId from the POM file's root element.

    Args:
        root (ET.Element): The root element of the POM file's XML tree.
        ns_mapping (dict): A dictionary mapping XML namespaces for parsing.

    Returns:
        str: The full artifact name in the format 'groupId:artifactId'.
            If groupId is not present, returns only the artifactId.
    """
    artifact = root.find('./xmlns:artifactId', namespaces=ns_mapping)
    artifact_text = str(artifact.text) if artifact is not None else ''
    group = root.find('./xmlns:groupId', namespaces=ns_mapping)
    return (str(group.text) + ':' if group is not None else '') + artifact_text


def collect_dependencies(
        root: ET.Element, ns_mapping: dict, config: Config, arguments: Arguments
) -> list:
    """
    Collects all dependency elements from the POM file.
    Optionally includes plugin elements if 'search_plugins' is enabled in the configuration.

    Args:
        root (ET.Element): The root element of the POM file's XML tree.
        ns_mapping (dict): A dictionary mapping XML namespaces for parsing.
        config (Config): Configuration dictionary parsed from YAML.
        arguments (Arguments): Command-line arguments.

    Returns:
        list[ET.Element]: A list of dependency elements (and plugin elements if specified).
    """
    dependencies = root.findall('.//xmlns:dependency', namespaces=ns_mapping)
    if _config.get_config_value(config, arguments, 'search_plugins'):
        plugins = root.findall('.//xmlns:plugins/xmlns:plugin', namespaces=ns_mapping)
        dependencies.extend(plugins)
    return dependencies


def get_dependency_identifiers(dependency: ET.Element, ns_mapping: dict) -> tuple[str, str]:
    """
    Extracts the groupId and artifactId from a dependency element.

    Args:
        dependency (ET.Element): The dependency XML element from the POM file.
        ns_mapping (dict): A dictionary mapping XML namespaces for parsing.

    Returns:
        tuple[str, str]: A tuple containing:
            - group (str): The groupId, or an empty string if not present.
            - artifact (str): The artifactId, or an empty string if not present.
    """
    artifact = dependency.find('xmlns:artifactId', namespaces=ns_mapping)
    group = dependency.find('xmlns:groupId', namespaces=ns_mapping)
    return (
        str(group.text) if group is not None else '',
        str(artifact.text) if artifact is not None else ''
    )


def fail_mode_if_required(
        config: Config, current_major_version: int, current_minor_version: int, item: str,
        major_version_threshold: int, minor_version_threshold: int, arguments: Arguments,
        version: Optional[str]
) -> None:
    """
    Checks if fail mode is enabled and if the version exceeds specified thresholds.
    Logs a warning and raises an AssertionError if thresholds are exceeded.

    Args:
        config (Config): Configuration dictionary parsed from YAML.
        current_major_version (int): The major version of the current artifact version.
        current_minor_version (int): The minor version of the current artifact version.
        item (str): The version string to check against thresholds (e.g., '1.2.3').
        major_version_threshold (int): The maximum allowed difference in major versions.
        minor_version_threshold (int): The maximum allowed difference in minor versions.
        arguments (Arguments): Command-line arguments.
        version (Optional[str]): The current version of the artifact (e.g., '1.0.0').
    """
    if _config.get_config_value(config, arguments, 'fail_mode'):
        item_major_version = 0
        item_minor_version = 0

        if item and (item_match := re.match(r'^(\d+).(\d+).?', item)):
            item_major_version, item_minor_version = int(item_match.group(1)), int(item_match.group(2))

        if item_major_version - current_major_version > major_version_threshold or \
                item_minor_version - current_minor_version > minor_version_threshold:
            logging.warning(f"Fail version: {item} > {version}")
            raise AssertionError


def resolve_version(version: str, root: ET.Element, ns_mapping: dict) -> str:
    """
    Resolves the version string by replacing placeholders with values from POM properties.
    Handles placeholders like '${property}' or '${project.version}'.

    Args:
        version (str): The version string, which may contain placeholders (e.g., '${version}').
        root (ET.Element): The root element of the POM file's XML tree.
        ns_mapping (dict): A dictionary mapping XML namespaces for parsing.

    Returns:
        str: The resolved version string if a placeholder is matched and found in properties,
            otherwise the original version string.
    """
    if match := re.match(r'^\${([^}]+)}$', version):
        property_xpath = f"./xmlns:properties/xmlns:{match.group(1)}"
        property_element = root.find(property_xpath, namespaces=ns_mapping)
        if property_element is not None:
            version = str(property_element.text)
    return version


def get_version(
        config: Config, arguments: Arguments, ns_mapping: dict, root: ET.Element,
        dependency: ET.Element
) -> tuple[Optional[str], bool]:
    """
    Extracts version information from a dependency.

    Args:
        config (Config): Parsed YAML as dict.
        arguments (Arguments): Command-line arguments.
        ns_mapping (dict): XML namespace mapping.
        root (ET.Element): Root element of the POM file.
        dependency (ET.Element): Dependency element.

    Returns:
        tuple[Optional[str], bool]: Tuple of version (or None) and skip flag.
    """
    version_text = ''
    version = dependency.find('xmlns:version', namespaces=ns_mapping)

    if version is None:
        if not _config.get_config_value(config, arguments, 'empty_version'):
            return None, True
    else:
        version_text = resolve_version(str(version.text), root, ns_mapping)

        if version_text == '${project.version}':
            project_version = root.find('xmlns:version', namespaces=ns_mapping)
            project_version_text = str(project_version.text) if project_version is not None else ''
            version_text = resolve_version(project_version_text, root, ns_mapping)

        if version_text and re.match(r'^\${([^}]+)}$', version_text):
            if not _config.get_config_value(config, arguments, 'empty_version'):
                return version_text, True

    return version_text, False


def check_versions(
        cache_data: Optional[dict], config: Config, arguments: Arguments, group: str, artifact: str,
        version: Optional[str], section_key: str, path: str, auth_info: Optional[tuple[str, str]],
        verify_ssl: bool, available_versions: list[str], response: requests.Response
) -> bool:
    """
    Checks dependency versions in a repository.

    Args:
        cache_data (Optional[dict]): Cache data.
        config (Config): Parsed YAML as dict.
        arguments (Arguments): Command-line arguments.
        group (str): Group ID.
        artifact (str): Artifact ID.
        version (Optional[str]): Current version.
        section_key (str): Repository section key.
        path (str): Path to the dependency in the repository.
        auth_info (Optional[tuple[str, str]]): Authentication credentials.
        verify_ssl (bool): SSL verification flag.
        available_versions (list[str]): List of available versions.
        response (requests.Response): Repository response.

    Returns:
        bool: True if the current version is valid, False otherwise.
    """
    major_threshold = minor_threshold = 0
    current_major = current_minor = 0

    if _config.get_config_value(config, arguments, 'fail_mode'):
        major_threshold = int(_config.get_config_value(config, arguments, 'fail_major'))
        minor_threshold = int(_config.get_config_value(config, arguments, 'fail_minor'))

        if version and (version_match := re.match(r'^(\d+)\.(\d+).?', version)):
            current_major, current_minor = int(version_match.group(1)), int(version_match.group(2))

    skip_current = _config.get_config_value(config, arguments, 'skip_current', default=True)
    invalid_flag = False

    for item in available_versions:
        if item == version and skip_current:
            _cache.update_cache_artifact(
                cache_data, available_versions, artifact, group, item, None, section_key)
            return True

        is_valid, last_modified = get_pom_data(auth_info, verify_ssl, artifact, item, path)
        if is_valid:
            logging.info('{}: {}:{}:{}, last versions: {}, modified:{}.'.format(
                section_key, group, artifact, version, available_versions[:5], last_modified).rstrip())

            _cache.update_cache_artifact(
                cache_data, available_versions, artifact, group, item, last_modified, section_key)

            fail_mode_if_required(
                config, current_major, current_minor, item,
                major_threshold, minor_threshold, arguments, version)
            return True

        else:
            _logutils.log_invalid_if_required(
                config, arguments, response, group, artifact, item, invalid_flag)
            invalid_flag = True

    return False


def get_pom_data(
        auth_info: Optional[tuple[str, str]], verify_ssl: bool, artifact: str, version: str, path: str
) -> tuple[bool, Optional[str]]:
    """
    Retrieves POM file data from a repository.

    Args:
        auth_info (Optional[tuple[str, str]]): Authentication credentials.
        verify_ssl (bool): SSL verification flag.
        artifact (str): Artifact ID.
        version (str): Artifact version.
        path (str): Path to the dependency in the repository.

    Returns:
        tuple[bool, Optional[str]]: Tuple of success flag and last modified date (or None).
    """
    with requests.Session() as session:
        url = f"{path}/{version}/{artifact}-{version}.pom"
        response = session.get(url, auth=auth_info, verify=verify_ssl)

        if response.status_code == 200:
            last_modified_header = response.headers.get('Last-Modified')
            return True, parser.parse(last_modified_header).date().isoformat()

    return False, None


def get_pom_tree(
        pom_path: str, verify_ssl: bool, config: Config, arguments: Arguments
) -> ET.ElementTree:
    """
    Loads the XML tree of a POM file.

    Args:
        pom_path (str): Path or URL to the POM file.
        verify_ssl (bool): SSL verification flag.
        config (Config): Parsed YAML as dict.
        arguments (Arguments): Command-line arguments.

    Returns:
        ET.ElementTree: Parsed XML tree of the POM file.
    """
    if pom_path.startswith('http'):
        auth_info: Optional[tuple[str, str]] = None
        if _config.get_config_value(config, arguments, 'auth', 'pom_http', default=False):
            auth_info = (
                _config.get_config_value(config, arguments, 'user', 'pom_http'),
                _config.get_config_value(config, arguments, 'password', 'pom_http')
            )
        with requests.Session() as session:
            response = session.get(pom_path, auth=auth_info, verify=verify_ssl)
            if response.status_code != 200:
                raise FileNotFoundError(f"Failed to get_pom_tree {pom_path}: HTTP {response.status_code}")
            return ET.ElementTree(ET.fromstring(response.text))
    else:
        if not os.path.exists(pom_path) or not os.path.isfile(pom_path):
            raise FileNotFoundError(f"Failed to get_pom_tree {pom_path}")
        return ET.parse(pom_path)



#!/usr/bin/python3
"""Tests for package process"""

import os
import sys
# noinspection PyPep8Naming
import xml.etree.ElementTree as ET
from typing import Optional

# noinspection PyUnresolvedReferences
from pytest_mock import mocker

os.chdir(os.path.dirname(__file__))
sys.path.append('../src')

# noinspection PyUnresolvedReferences
from maven_check_versions.process import (  # noqa: E402
    process_rest, process_repository, process_repositories,
    process_modules_if_required, process_artifact,
    process_dependency, process_pom, process_main
)

# noinspection PyUnresolvedReferences
from maven_check_versions.logutils import (  # noqa: E402
    configure_logging, log_skip_if_required,
    log_search_if_required, log_invalid_if_required
)

# noinspection PyUnresolvedReferences
from maven_check_versions.config import (  # noqa: E402
    get_config_value, config_items, Arguments, Config
)

# noinspection PyUnresolvedReferences
from maven_check_versions.utils import (  # noqa: E402
    get_dependency_identifiers, collect_dependencies,
    resolve_version, get_version,
)

ns_mappings = {'xmlns': 'http://maven.apache.org/POM/4.0.0'}  # NOSONAR


# noinspection PyShadowingNames
def test_process_main(mocker, monkeypatch):
    monkeypatch.setenv('HOME', os.path.dirname(__file__))
    mock_exists = mocker.patch('os.path.exists')
    mock_exists.side_effect = [False, False, True]
    mocker.patch('builtins.open', mocker.mock_open(read_data="base.cache_off: false"))
    mocker.patch('maven_check_versions.cache.load_cache', return_value={})
    mocker.patch('maven_check_versions.process.process_pom')
    mocker.patch('maven_check_versions.cache.save_cache')
    process_main(Arguments({'pom_file': 'pom.xml'}))

    mock_exists.side_effect = [False, False, True]
    mocker.patch('maven_check_versions.process.process_artifact')
    process_main(Arguments({'find_artifact': 'pom.xml'}))

    mock_exists.side_effect = [False, False, True]
    mock_config_items = mocker.patch('maven_check_versions.config.config_items')
    mock_config_items.return_value = [('key', 'pom.xml')]
    process_main(Arguments())


# noinspection PyShadowingNames
def test_process_rest(mocker):
    def _service_rest():
        return process_rest(
            {}, Config(), Arguments(), 'group', 'artifact', '1.0', 'section',
            'repository', 'http://example.com/pom.pom', None, True  # NOSONAR
        )

    mock_check_versions = mocker.patch('maven_check_versions.utils.check_versions')
    mock_check_versions.return_value = True
    mock_requests = mocker.patch('requests.Session.get')
    mock_requests.return_value = mocker.Mock(status_code=200, text="""
    <?xml version="1.0" encoding="UTF-8"?>
    <root>
        <version>1.0</version>
        <version>1.1</version>
    </root>
    """.lstrip())
    assert _service_rest()

    text = '<html><body><table><a>1.0</a><a>1.1</a></table></body></html>'
    mock_response = mocker.Mock(status_code=200, text=text)
    mock_requests.side_effect = [mocker.Mock(status_code=404), mock_response]
    assert _service_rest()

    mock_response = mocker.Mock(status_code=404)
    mock_requests.side_effect = [mock_response, mock_response]
    assert not _service_rest()


# noinspection PyShadowingNames
def test_process_repository(mocker):
    config = Config({
        'section': {
            'base': 'https://repo1.maven.org',
            'path': 'maven2',
            'repo': 'maven-central',
            'use_rest': True,
            'auth': True,
        }})
    args = Arguments({'user': 'user', 'password': 'pass'})  # NOSONAR

    def _process_repository():
        return process_repository(
            {}, config, args, 'group', 'artifact', '1.0',
            'repository', 'section', True
        )

    mock_requests = mocker.patch('requests.Session.get')
    mock_requests.return_value = mocker.Mock(status_code=200, text="""
    <?xml version="1.0" encoding="UTF-8"?>
    <metadata>
        <versioning>
            <versions>
                <version>1.0</version>
                <version>1.1</version>
            </versions>
        </versioning>
    </metadata>
    """.lstrip())
    mock_check_versions = mocker.patch('maven_check_versions.utils.check_versions')
    mock_check_versions.return_value = True
    assert _process_repository()

    mock_requests.return_value = mocker.Mock(status_code=404)
    mock_process_rest = mocker.patch('maven_check_versions.process.process_rest')
    mock_process_rest.return_value = True
    assert _process_repository()

    config['section']['use_rest'] = False
    assert not _process_repository()


# noinspection PyShadowingNames
def test_process_repositories(mocker):
    config = Config({
        'repositories': {
            'repo1': 'maven-central',
            'repo2': 'custom-repo',
        },
        'maven-central': {
            'base': 'https://repo1.maven.org',
            'path': 'maven2',
        },
        'custom-repo': {
            'base': 'https://custom.repo',
            'path': 'maven2',
        }
    })
    mock_process_repository = mocker.patch('maven_check_versions.process.process_repository')
    mock_process_repository.return_value = True
    assert process_repositories('artifact', {}, config, 'group', Arguments(), True, '1.0')

    config = Config({'repositories': {}})
    assert not process_repositories('artifact', {}, config, 'group', Arguments(), True, '1.0')


# noinspection PyShadowingNames
def test_process_modules_if_required(mocker):
    config = Config({
        'base': {
            'process_modules': True,
            'threading': True,
            'max_threads': 1
        }
    })
    root = ET.fromstring("""
    <?xml version="1.0" encoding="UTF-8"?>
    <project xmlns="http://maven.apache.org/POM/4.0.0">
        <modules>
            <module>module1</module>
            <module>module2</module>
        </modules>
    </project>
    """.lstrip())
    mocker.patch('os.path.exists', return_value=True)
    mock_process_pom = mocker.patch('maven_check_versions.process.process_pom')
    process_modules_if_required({}, config, Arguments(), root, 'pom.xml', ns_mappings)
    assert mock_process_pom.call_count == 2

    config['base']['threading'] = False
    mock_process_pom = mocker.patch('maven_check_versions.process.process_pom')
    process_modules_if_required({}, config, Arguments(), root, 'pom.xml', ns_mappings)
    assert mock_process_pom.call_count == 2


# noinspection PyShadowingNames
def test_process_artifact(mocker):
    config = Config({
        'base': {
            'show_search': 'true',
        },
        'repositories': {
            'repo1': 'maven-central',
            'repo2': 'custom-repo',
        },
        'maven-central': {
            'base': 'https://repo1.maven.org',
            'path': 'maven2',
        },
        'custom-repo': {
            'base': 'https://custom.repo',
            'path': 'maven2',
        }
    })
    mock_logging = mocker.patch('logging.info')
    mock_process_repository = mocker.patch('maven_check_versions.process.process_repository')
    mock_process_repository.return_value = True
    process_artifact(None, config, Arguments(), 'group:artifact:1.0')
    mock_logging.assert_called_once_with('Search: group:artifact:1.0')
    mock_process_repository.assert_called_once()

    mock_logging = mocker.patch('logging.warning')
    mock_process_repository.return_value = False
    process_artifact(None, config, Arguments(), 'group:artifact:1.0')
    mock_logging.assert_called_once_with('Not Found: group:artifact, current:1.0')


# noinspection PyShadowingNames
def test_process_dependency(mocker):
    config = Config({
        'base': {
            'empty_version': 'true',
            'show_skip': 'true',
        }
    })
    root = ET.fromstring("""
        <?xml version="1.0" encoding="UTF-8"?>
        <project xmlns="http://maven.apache.org/POM/4.0.0">
            <dependencies>
                <dependency xmlns="http://maven.apache.org/POM/4.0.0">
                    <artifactId>artifact</artifactId>
                    <groupId>group</groupId>
                    <version>1.0</version>
                </dependency>
            </dependencies>
        </project>
        """.lstrip())
    dependencies = collect_dependencies(root, ns_mappings, config, Arguments())

    def _process_dependencies(data: Optional[dict] = None) -> None:
        for dep in dependencies:
            process_dependency(data, config, Arguments(), dep, ns_mappings, root, True)

    mock_gdi = mocker.patch('maven_check_versions.utils.get_dependency_identifiers')
    mock_gdi.return_value = ('artifact', None)
    mock_logging = mocker.patch('logging.error')
    _process_dependencies()
    mock_logging.assert_called_once()

    mock_gdi.return_value = ('group', 'artifact')
    mock_get_version = mocker.patch('maven_check_versions.utils.get_version')
    mock_get_version.return_value = ('1.0', True)
    mock_logging = mocker.patch('logging.warning')
    _process_dependencies()
    mock_logging.assert_called_once()

    mock_get_version.return_value = ('1.0', False)
    mocker.patch('maven_check_versions.cache.process_cache_artifact', return_value=True)
    _process_dependencies({'group:artifact': ()})

    mocker.patch('maven_check_versions.process.process_repositories', return_value=False)
    mock_logging = mocker.patch('logging.warning')
    _process_dependencies()
    mock_logging.assert_called_once()


# noinspection PyShadowingNames
def test_process_pom(mocker):
    mock_get_pom_tree = mocker.patch('maven_check_versions.utils.get_pom_tree')
    mock_get_pom_tree.return_value = ET.ElementTree(ET.fromstring("""
    <project xmlns="http://maven.apache.org/POM/4.0.0">
        <artifactId>artifact</artifactId>
        <groupId>group</groupId>
        <version>1.0</version>
        <dependencies>
            <dependency>
                <artifactId>artifact</artifactId>
                <groupId>group</groupId>
                <version>1.0</version>
            </dependency>
        </dependencies>
    </project>
    """))
    mock_pd = mocker.patch('maven_check_versions.process.process_dependency')
    mock_pm = mocker.patch('maven_check_versions.process.process_modules_if_required')
    config = Config({
        'base': {
            'threading': True,
            'max_threads': 4
        }
    })
    process_pom({}, config, Arguments(), 'pom.xml', 'prefix')
    mock_get_pom_tree.assert_called_once()
    mock_pd.assert_called_once()
    mock_pm.assert_called_once()

    config['base']['threading'] = False
    mock_pd = mocker.patch('maven_check_versions.process.process_dependency')
    process_pom({}, config, Arguments(), 'pom.xml', 'prefix')
    mock_pd.assert_called_once()



#!/usr/bin/python3
"""Tests for package config functions"""

import os
import sys

# noinspection PyUnresolvedReferences
from pytest_mock import mocker

os.chdir(os.path.dirname(__file__))
sys.path.append('../src')

from maven_check_versions.config import (
    get_config, get_config_value, config_items, Config, Arguments
)


# noinspection PyShadowingNames
def test_get_config(mocker):
    mock_exists = mocker.patch('os.path.exists')
    mock_exists.side_effect = [False, True]
    mocker.patch('builtins.open', mocker.mock_open(read_data="base:"))
    mock_logging = mocker.patch('logging.info')
    get_config(Arguments())
    mock_logging.assert_called_once()
    mocker.stopall()


# noinspection PyShadowingNames
def test_get_config_value(monkeypatch):
    config = Config({'base': {'key': True}, 'other': {'key': False}})
    assert get_config_value(config, Arguments(), 'key') is True
    assert get_config_value(config, Arguments(), 'val', default=True) is True
    assert get_config_value(config, Arguments({'key': False}), 'key') is False
    assert get_config_value(config, Arguments(), 'val') is None
    monkeypatch.setenv('CV_KEY', 'false')
    assert get_config_value(config, Arguments({'key': None}), 'key') is False
    monkeypatch.setenv('CV_KEY', 'true')
    assert get_config_value(config, Arguments({'key': None}), 'key') is True
    assert get_config_value(config, Arguments({'key': False}), 'key') is False
    monkeypatch.setenv('CV_OTHER_KEY', 'true')
    assert get_config_value(config, Arguments(), 'key', section='other') is True
    monkeypatch.undo()
    config = Config({'base': {'key': 123}})
    assert get_config_value(config, Arguments(), 'key') == 123
    assert get_config_value(config, Arguments(), 'val', default=123) == 123
    config = Config({'base': {'key': 123.45}})
    assert get_config_value(config, Arguments(), 'key') == 123.45  # NOSONAR
    assert get_config_value(config, Arguments(), 'val', default=123.45) == 123.45  # NOSONAR
    config = Config({'base': {'key': 'value'}})
    assert get_config_value(config, Arguments(), 'key') == 'value'
    assert get_config_value(config, Arguments(), 'val', default='value') == 'value'
    assert get_config_value(config, Arguments(), 'val', default=[]) == []
    assert get_config_value(config, Arguments(), 'val', default=()) == ()
    assert get_config_value(config, Arguments(), 'val', default={'k': 'v'}) == {'k': 'v'}


def test_config_items():
    config = Config({'base': {'key': 'value'}})
    assert config_items(config, 'base') == [('key', 'value')]
    assert config_items(config, 'other') == []
    assert config_items(config, 'empty') == []



#!/usr/bin/python3
"""This file provides config functions"""

import logging
import os
from pathlib import Path
from typing import Any

import yaml


class Config(dict):
    """Wrapper for Config"""
    pass


class Arguments(dict):
    """Wrapper for Arguments"""
    pass


def get_config(arguments: Arguments) -> Config:
    """
    Loads the configuration from a YAML file specified in the arguments or a default location.

    Args:
        arguments (Arguments): Command-line arguments.

    Returns:
        Config: A Config object (dictionary) containing the parsed YAML configuration.
                If no config file is found, returns an empty Config object.
    """
    config = Config()
    if (config_file := arguments.get('config_file')) is None:
        config_file = 'maven_check_versions.yml'
        if not os.path.exists(config_file):
            config_file = os.path.join(Path.home(), config_file)

    if os.path.exists(config_file):
        logging.info(f"Load Config: {Path(config_file).absolute()}")
        try:
            with open(config_file, encoding='utf-8') as f:
                config = yaml.safe_load(f)
        except Exception as e:  # pragma: no cover
            logging.error(f"Failed to get_config: {e}")

    return config


def get_config_value(
        config: Config, arguments: Arguments, key: str, section: str = 'base', default: Any = None
) -> Any:
    """
    Retrieves a configuration value from command-line arguments or the config dictionary,
    with a fallback to a default value.

    Args:
        config (Config): Configuration dictionary parsed from YAML.
        arguments (Arguments): Command-line arguments.
        key (str): Configuration key.
        section (str, optional): Configuration section to use (default is 'base').
        default (Any, optional): Default value if the key is not found (default is None).

    Returns:
        Any: The value associated with the key, sourced from arguments, environment variables,
            or the config dictionary, or the default value if not found.
    """
    value = None
    if section == 'base' and key in arguments:
        value = arguments.get(key)
    if value is None:
        env = 'CV_' + (('' if section == 'base' else section + '_') + key).upper()
        if env in os.environ and (value := os.environ.get(env)) is not None:
            if value.lower() == 'true':
                value = True
            elif value.lower() == 'false':
                value = False
    if value is None and section in config and (get := config.get(section)):
        value = get.get(key)
    return default if value is None else value


def config_items(config: Config, section: str) -> list[tuple[str, str]]:
    """
    Retrieves all key-value pairs from a specified configuration section.

    Args:
        config (Config): Configuration dictionary parsed from YAML.
        section (str): The name of the configuration section (e.g., 'repositories').

    Returns:
        list[tuple[str, str]]: A list of tuples, each containing a key and its value from the section.
                            Returns an empty list if the section does not exist.
    """
    return list(get.items()) if (get := config.get(section)) else []



#!/usr/bin/python3
"""Tests for package utility functions"""

import os
import sys
# noinspection PyPep8Naming
import xml.etree.ElementTree as ET

import pytest
# noinspection PyUnresolvedReferences
from pytest_mock import mocker

os.chdir(os.path.dirname(__file__))
sys.path.append('../src')

# noinspection PyUnresolvedReferences
from maven_check_versions.utils import (  # noqa: E402
    parse_command_line, get_artifact_name, collect_dependencies,
    get_dependency_identifiers, fail_mode_if_required, resolve_version,
    get_version, check_versions, get_pom_data, get_pom_tree
)
from maven_check_versions.config import Arguments, Config

ns_mappings = {'xmlns': 'http://maven.apache.org/POM/4.0.0'}  # NOSONAR


# noinspection PyShadowingNames
def test_parse_command_line(mocker):
    mocker.patch(
        'argparse.ArgumentParser.parse_args',
        return_value=mocker.Mock(
            ci_mode=True,
            pom_file='pom.xml',
            find_artifact='artifact',
            cache_off=True,
            cache_file='cache.json',
            cache_time=3600,
            logfile_off=True,
            log_file='log.txt',
            config_file='config.cfg',
            fail_mode=True,
            fail_major=1,
            fail_minor=2,
            search_plugins=True,
            process_modules=True,
            show_skip=True,
            show_search=True,
            empty_version=True,
            show_invalid=True,
            user='user',
            password='password'
        ))
    args = parse_command_line()
    assert args['ci_mode'] is True
    assert args['pom_file'] == 'pom.xml'
    assert args['find_artifact'] == 'artifact'
    assert args['cache_off'] is True
    assert args['cache_file'] == 'cache.json'
    assert args['cache_time'] == 3600
    assert args['logfile_off'] is True
    assert args['log_file'] == 'log.txt'
    assert args['config_file'] == 'config.cfg'
    assert args['fail_mode'] is True
    assert args['fail_major'] == 1
    assert args['fail_minor'] == 2
    assert args['search_plugins'] is True
    assert args['process_modules'] is True
    assert args['show_skip'] is True
    assert args['show_search'] is True
    assert args['empty_version'] is True
    assert args['show_invalid'] is True
    assert args['user'] == 'user'
    assert args['password'] == 'password'


def test_get_artifact_name():
    root = ET.fromstring("""
    <?xml version="1.0" encoding="UTF-8"?>
    <project xmlns="http://maven.apache.org/POM/4.0.0">
        <groupId>groupId</groupId>
        <artifactId>artifactId</artifactId>
        <version>1.0</version>
    </project>
    """.lstrip())
    result = get_artifact_name(root, ns_mappings)
    assert result == "groupId:artifactId"


# noinspection PyShadowingNames
def test_collect_dependencies(mocker):
    root = ET.fromstring("""
    <?xml version="1.0" encoding="UTF-8"?>
    <project xmlns="http://maven.apache.org/POM/4.0.0">
        <dependencies>
            <dependency>
                <groupId>groupId</groupId>
                <artifactId>artifactId</artifactId>
            </dependency>
            <dependency>
                <groupId>groupId</groupId>
                <artifactId>artifactId</artifactId>
            </dependency>
        </dependencies>
        <build>
            <plugins>
            <plugin>
                <groupId>groupId</groupId>
                <artifactId>artifactId</artifactId>
            </plugin>
            </plugins>
        </build>
    </project>
    """.lstrip())
    args = Arguments({'search_plugins': True})
    result = collect_dependencies(root, ns_mappings, Config(), args)
    assert len(result) == 3


def test_get_dependency_identifiers():
    dependency = ET.fromstring("""
    <?xml version="1.0" encoding="UTF-8"?>
    <dependency xmlns="http://maven.apache.org/POM/4.0.0">
        <groupId>groupId</groupId>
        <artifactId>artifactId</artifactId>
        <version>1.0</version>
    </dependency>
    """.lstrip())
    group, artifact = get_dependency_identifiers(dependency, ns_mappings)
    assert group == 'groupId' and artifact == 'artifactId'


# noinspection PyShadowingNames
def test_fail_mode_if_required(mocker):
    mock_logging = mocker.patch('logging.warning')
    with pytest.raises(AssertionError):
        config = Config()
        args = Arguments({'fail_mode': True, 'fail_major': 2, 'fail_minor': 2})
        fail_mode_if_required(config, 1, 0, '4.0', 2, 2, args, '1.0')
    mock_logging.assert_called_once_with("Fail version: 4.0 > 1.0")


def test_resolve_version():
    root = ET.fromstring("""
    <?xml version="1.0" encoding="UTF-8"?>
    <project xmlns="http://maven.apache.org/POM/4.0.0">
        <properties>
            <lib.version>1.0</lib.version>
        </properties>
    </project>
    """.lstrip())
    version = resolve_version('${lib.version}', root, ns_mappings)
    assert version == '1.0'


def test_get_version():
    root = ET.fromstring("""
    <?xml version="1.0" encoding="UTF-8"?>
    <project xmlns="http://maven.apache.org/POM/4.0.0">
        <version>1.0</version>
        <dependencies>
            <dependency>
                <artifactId>dependency</artifactId>
            </dependency>
            <dependency>
                <artifactId>dependency</artifactId>
                <version>${project.version}</version>
            </dependency>
            <dependency>
                <artifactId>dependency</artifactId>
                <version>${dependency.version}</version>
            </dependency>
        </dependencies>
    </project>
    """.lstrip())
    args = Arguments({'empty_version': False})
    deps = root.findall('.//xmlns:dependency', namespaces=ns_mappings)
    version, skip_flag = get_version(Config(), args, ns_mappings, root, deps[0])
    assert version is None and skip_flag

    version, skip_flag = get_version(Config(), args, ns_mappings, root, deps[1])
    assert version == '1.0' and not skip_flag

    version, skip_flag = get_version(Config(), args, ns_mappings, root, deps[2])
    assert version == '${dependency.version}' and skip_flag


# noinspection PyShadowingNames
def test_check_versions(mocker):
    def _check_versions(pa, data, item, vers):
        return check_versions(
            data, Config(), pa, 'group', 'artifact', item,
            'repo_section', 'path', None, True, vers, mocker.Mock()
        )

    mock_get_pom_data = mocker.patch('maven_check_versions.utils.get_pom_data')
    mock_get_pom_data.return_value = (True, '2025-01-25')
    args = Arguments({
        'skip_current': True, 'fail_mode': True,
        'fail_major': 0, 'fail_minor': 1
    })
    cache_data = {}
    assert _check_versions(args, cache_data, '1.1', ['1.1'])
    assert cache_data['group:artifact'][1] == '1.1'

    with pytest.raises(AssertionError):
        args['fail_minor'] = 0
        assert _check_versions(args, cache_data, '1.1', ['1.2'])

    args['fail_mode'] = False
    assert _check_versions(args, cache_data, '1.1', ['1.2'])

    mock_get_pom_data.return_value = (False, None)
    assert not _check_versions(args, cache_data, '1.1', ['1.2'])


# noinspection PyShadowingNames
def test_get_pom_data(mocker):
    pom_path = 'http://example.com/pom.pom'  # NOSONAR
    headers = {'Last-Modified': 'Wed, 18 Jan 2025 12:00:00 GMT'}
    mock_response = mocker.Mock(status_code=200, headers=headers)
    mock_requests = mocker.patch('requests.Session.get', return_value=mock_response)
    is_valid, last_modified = get_pom_data(None, True, 'artifact', '1.0', pom_path)
    assert is_valid is True and last_modified == '2025-01-18'

    mock_requests.return_value = mocker.Mock(status_code=404)
    is_valid, last_modified = get_pom_data(None, True, 'artifact', '1.0', pom_path)
    assert is_valid is False and last_modified is None


# noinspection PyShadowingNames
def test_get_pom_tree(mocker):
    xml = """<?xml version="1.0" encoding="UTF-8"?>
    <project xmlns="http://maven.apache.org/POM/4.0.0">
        <groupId>group</groupId>
        <artifactId>artifact</artifactId>
        <version>1.0</version>
    </project>
    """
    config = Config({'pom_http': {'auth': 'true'}})
    mocker.patch('os.path.exists', return_value=True)
    mocker.patch('os.path.isfile', return_value=True)
    mock_open = mocker.patch('builtins.open', mocker.mock_open(read_data=xml))
    tree = get_pom_tree('pom.xml', True, config, Arguments())
    mock_open.assert_called_once_with('pom.xml', 'rb')
    assert isinstance(tree, ET.ElementTree)

    mocker.patch('os.path.exists', return_value=False)
    with pytest.raises(FileNotFoundError):
        get_pom_tree('pom.xml', True, config, Arguments())

    pom_path = 'http://example.com/pom.pom'  # NOSONAR
    mock_response = mocker.Mock(status_code=200, text=xml)
    mock_requests = mocker.patch('requests.Session.get', return_value=mock_response)
    assert isinstance(get_pom_tree(pom_path, True, config, Arguments()), ET.ElementTree)

    mock_requests.return_value.status_code = 404
    with pytest.raises(FileNotFoundError):
        get_pom_tree(pom_path, True, config, Arguments())



#!/usr/bin/python3
"""Tests for package cache functions"""

import os
import sys
import time
from json import JSONDecodeError

import pytest
# noinspection PyUnresolvedReferences
from pytest_mock import mocker

os.chdir(os.path.dirname(__file__))
sys.path.append('../src')

from maven_check_versions.config import Config, Arguments
from maven_check_versions.cache import (
    load_cache, save_cache, update_cache_artifact,
    process_cache_artifact, DCJSONEncoder
)


# noinspection PyShadowingNames
def test_load_cache(mocker):
    mocker.patch('os.path.exists', return_value=True)
    mocker.patch('builtins.open', mocker.mock_open(read_data='{"k": "v"}'))
    assert load_cache(Config(), Arguments()) == {'k': 'v'}

    mocker.patch('json.load').side_effect = JSONDecodeError('error', 'error', 0)
    assert load_cache(Config(), Arguments()) == {}

    mocker.patch('os.path.exists', return_value=False)
    assert load_cache(Config(), Arguments()) == {}

    mock_redis = mocker.patch('redis.Redis')
    mock_redis.return_value.hgetall.return_value = {'key': '{"k":"v"}'}
    assert load_cache(Config({'base': {'cache_backend': 'redis'}}), Arguments()) == {'key': {'k': 'v'}}

    mock_loads = mocker.patch('json.loads')
    mock_loads.side_effect = JSONDecodeError('error', 'error', 0)
    assert load_cache(Config({'base': {'cache_backend': 'redis'}}), Arguments()) == {}
    mocker.stop(mock_loads)

    mock_redis.side_effect = Exception
    assert load_cache(Config({'base': {'cache_backend': 'redis'}}), Arguments()) == {}

    mock_tarantool = mocker.patch('tarantool.Connection')
    mock_tarantool.return_value.select.return_value = [('key', '{"k":"v"}')]
    assert load_cache(Config({'base': {'cache_backend': 'tarantool'}}), Arguments()) == {'key': {'k': 'v'}}

    mock_loads = mocker.patch('json.loads')
    mock_loads.side_effect = JSONDecodeError('error', 'error', 0)
    assert load_cache(Config({'base': {'cache_backend': 'tarantool'}}), Arguments()) == {}
    mocker.stop(mock_loads)

    mock_tarantool.side_effect = Exception
    assert load_cache(Config({'base': {'cache_backend': 'tarantool'}}), Arguments()) == {}

    mock_memcache = mocker.patch('pymemcache.client.base.Client')
    mock_memcache.return_value.get.return_value = '{"k":"v"}'
    assert load_cache(Config({'base': {'cache_backend': 'memcached'}}), Arguments()) == {'k': 'v'}

    mock_loads = mocker.patch('json.loads')
    mock_loads.side_effect = JSONDecodeError('error', 'error', 0)
    assert load_cache(Config({'base': {'cache_backend': 'memcached'}}), Arguments()) == {}
    mocker.stop(mock_loads)

    mock_memcache.side_effect = Exception
    assert load_cache(Config({'base': {'cache_backend': 'memcached'}}), Arguments()) == {}
    mocker.stopall()


# noinspection PyShadowingNames
def test_save_cache(mocker):
    mock_open = mocker.patch('builtins.open')
    mock_json = mocker.patch('json.dumps')
    save_cache(Config(), Arguments(), {'k': 'v'})
    mock_open.assert_called_once_with('cache_maven_check_versions_artifacts.json', 'w', encoding='utf-8')
    mock_json.assert_called_once_with({'k': 'v'}, cls=DCJSONEncoder, indent=2)

    mock_json.side_effect = Exception
    save_cache(Config(), Arguments(), {'k': 'v'})

    mock_redis = mocker.patch('redis.Redis')
    save_cache(Config({'base': {'cache_backend': 'redis'}}), Arguments(), {'k': 'v'})

    mock_redis.side_effect = Exception
    save_cache(Config({'base': {'cache_backend': 'redis'}}), Arguments(), {'k': 'v'})

    mock_tarantool = mocker.patch('tarantool.Connection')
    save_cache(Config({'base': {'cache_backend': 'tarantool'}}), Arguments(), {'k': 'v'})

    mock_tarantool.side_effect = Exception
    save_cache(Config({'base': {'cache_backend': 'tarantool'}}), Arguments(), {'k': 'v'})

    mock_memcache = mocker.patch('pymemcache.client.base.Client')
    save_cache(Config({'base': {'cache_backend': 'memcached'}}), Arguments(), {'k': 'v'})

    mock_memcache.side_effect = Exception
    save_cache(Config({'base': {'cache_backend': 'memcached'}}), Arguments(), {'k': 'v'})
    mocker.stopall()


# noinspection PyShadowingNames
def test_process_cache_artifact(mocker):
    config = Config()
    data = {'group:artifact': (time.time() - 100, '1.0', 'key', '23.01.2025', ['1.0', '1.1'])}
    assert process_cache_artifact(config, Arguments({'cache_time': 0}), data, 'artifact', 'group', '1.0')
    assert not process_cache_artifact(config, Arguments({'cache_time': 50}), data, 'artifact', 'group', '1.1')

    mock = mocker.patch('logging.info')
    assert process_cache_artifact(config, Arguments({'cache_time': 0}), data, 'artifact', 'group', '1.1')
    mock.assert_called_once_with('*key: group:artifact:1.1, last versions: 1.0, 1.1, modified:23.01.2025.')

    assert not process_cache_artifact(config, Arguments(), {}, 'artifact', 'group', '1.1')


def test_update_cache_artifact():
    cache_data = {}
    update_cache_artifact(cache_data, ['1.0'], 'artifact', 'group', '1.0', '16.01.2025', 'key')  # NOSONAR
    data = (pytest.approx(time.time()), '1.0', 'key', '16.01.2025', ['1.0'])
    assert cache_data == {'group:artifact': data}



#!/usr/bin/python3
"""Tests for package cve check functions"""
import os
import sys
# noinspection PyPep8Naming
import xml.etree.ElementTree as ET

import pytest
# noinspection PyUnresolvedReferences
from pytest_mock import mocker

os.chdir(os.path.dirname(__file__))
sys.path.append('../src')

from maven_check_versions.config import Config, Arguments
from maven_check_versions.cveutils import Vulnerability, log_vulnerability, get_cve_data
from maven_check_versions.utils import collect_dependencies


# noinspection PyShadowingNames
def test_log_vulnerability(mocker):
    cves = {'pkg:maven/group/artifact@1.0': [Vulnerability(id='1', cvssScore=1)]}
    mock_logging = mocker.patch('logging.warning')
    config = Config({'vulnerability': {'fail_score': 2.0, 'cve_reference': True}})
    log_vulnerability(config, Arguments(), 'group', 'artifact', '1.0', cves)
    msg = 'Vulnerability for group:artifact:1.0: cvssScore=1 cve=None cwe=None None None'
    mock_logging.assert_called_once_with(msg)

    with pytest.raises(AssertionError):
        mock_logging = mocker.patch('logging.error')
        config = Config({'vulnerability': {'fail_score': 1.0}})
        log_vulnerability(config, Arguments(), 'group', 'artifact', '1.0', cves)
        mock_logging.assert_called_once_with('cvssScore=1 >= fail-score=1')


# noinspection PyShadowingNames
def test_get_cve_data(mocker):
    root = ET.fromstring("""
    <?xml version="1.0" encoding="UTF-8"?>
    <project xmlns="http://maven.apache.org/POM/4.0.0">
        <dependencies>
            <dependency>
                <groupId>group1</groupId>
                <artifactId>artifact</artifactId>
                <version>1.0</version>
            </dependency>
            <dependency>
                <groupId>group2</groupId>
                <artifactId>artifact</artifactId>
                <version>1.0</version>
            </dependency>
            <dependency>
                <groupId>group3</groupId>
                <artifactId>artifact</artifactId>
                <version>${version}</version>
            </dependency>
            <dependency>
                <groupId>group4</groupId>
                <artifactId>artifact</artifactId>
                <version>1.0</version>
            </dependency>
        </dependencies>
    </project>
    """.lstrip())
    ns_mappings = {'xmlns': 'http://maven.apache.org/POM/4.0.0'}  # NOSONAR
    config = Config({'vulnerability': {
        'oss_index_enabled': True, 'skip_no_versions': True,
        'skip_checks': ['group2.*']
    }})
    deps = collect_dependencies(root, ns_mappings, config, Arguments())
    mock_load_cache = mocker.patch('maven_check_versions.cache.load_cache')
    mock_load_cache.return_value = {'pkg:maven/group4/artifact@1.0': []}
    mocker.patch('maven_check_versions.cache.save_cache')
    mock_requests = mocker.patch(
        'requests.Session.post',
        return_value=mocker.Mock(status_code=200, json=lambda: [{
            'coordinates': 'pkg:maven/group1/artifact@1.0',
            'vulnerabilities': [{'id': '1', 'cvssScore': 1}]
        }]))
    assert get_cve_data(config, Arguments(), deps, root, ns_mappings) == {
        'pkg:maven/group1/artifact@1.0': [Vulnerability(id='1', cvssScore=1)],
        'pkg:maven/group4/artifact@1.0': []
    }

    mock_load_cache.return_value = {}
    mock_requests.return_value.status_code = 404
    assert get_cve_data(config, Arguments(), deps, root, ns_mappings) == {}

    mock_requests.return_value = Exception()
    assert get_cve_data(config, Arguments(), deps, root, ns_mappings) == {}



#!/usr/bin/python3
"""This file provides process functions"""

import logging
import os
# noinspection PyPep8Naming
import xml.etree.ElementTree as ET
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Optional

import maven_check_versions.cache as _cache
import maven_check_versions.config as _config
import maven_check_versions.cveutils as _cveutils
import maven_check_versions.logutils as _logutils
import maven_check_versions.utils as _utils
import requests
import urllib3
from bs4 import BeautifulSoup
from maven_check_versions.config import Config, Arguments
from maven_check_versions.cveutils import Vulnerability


def process_main(arguments: Arguments) -> None:
    """
    Orchestrates the main processing logic for checking Maven dependencies.

    Args:
        arguments (Arguments): Command-line arguments.
                May specify 'pom_file', 'find_artifact', or rely on config for POM files.
    """
    config = _config.get_config(arguments)

    if not _config.get_config_value(config, arguments, 'warnings', 'urllib3'):
        urllib3.disable_warnings()

    cache_disabled = _config.get_config_value(config, arguments, 'cache_off')
    cache_data = _cache.load_cache(config, arguments) if not cache_disabled else None

    if pom_file := arguments.get('pom_file'):
        process_pom(cache_data, config, arguments, pom_file)
    elif artifact_to_find := arguments.get('find_artifact'):
        process_artifact(cache_data, config, arguments, artifact_to_find)
    else:
        for _, pom in _config.config_items(config, 'pom_files'):
            process_pom(cache_data, config, arguments, pom)

    _cache.save_cache(config, arguments, cache_data)


def process_pom(
        cache_data: Optional[dict], config: Config, arguments: Arguments,
        pom_path: str, prefix: Optional[str] = None
) -> None:
    """
    Processes a single POM file by extracting dependencies, checking versions,
    and optionally processing modules and vulnerabilities.

    Args:
        cache_data (Optional[dict]): Cache data.
        config (Config): Parsed YAML as dict.
        arguments (Arguments): Command-line arguments.
        pom_path (str): Local path or URL to the POM file to process.
        prefix (str, optional): Prefix to prepend to the artifact name in logs (default is None).
    """
    verify_ssl = _config.get_config_value(config, arguments, 'verify', 'requests')

    tree = _utils.get_pom_tree(pom_path, verify_ssl, config, arguments)
    root = tree.getroot()
    ns_mapping = {'xmlns': 'http://maven.apache.org/POM/4.0.0'}  # NOSONAR

    artifact_name = _utils.get_artifact_name(root, ns_mapping)
    if prefix is not None:
        artifact_name = f"{prefix} / {artifact_name}"
    logging.info(f"=== Processing: {artifact_name} ===")

    dependencies = _utils.collect_dependencies(root, ns_mapping, config, arguments)

    cve_data = _cveutils.get_cve_data(config, arguments, dependencies, root, ns_mapping)

    if _config.get_config_value(config, arguments, 'threading'):
        max_threads = _config.get_config_value(config, arguments, 'max_threads')

        with ThreadPoolExecutor(max_workers=max_threads) as executor:
            for future in as_completed([
                executor.submit(process_dependency,
                                cache_data, config, arguments, dep, ns_mapping, root, verify_ssl, cve_data)
                for dep in dependencies
            ]):
                try:
                    future.result()
                except Exception as e:  # pragma: no cover
                    logging.error(f"Error processing dependency: {e}")
    else:
        for dep in dependencies:
            process_dependency(cache_data, config, arguments, dep, ns_mapping, root, verify_ssl, cve_data)

    process_modules_if_required(cache_data, config, arguments, root, pom_path, ns_mapping, artifact_name)


def process_dependency(
        cache_data: Optional[dict], config: Config, arguments: Arguments, dependency: ET.Element, ns_mapping: dict,
        root: ET.Element, verify_ssl: bool, cve_data: Optional[dict[str, list[Vulnerability]]] = None
) -> None:
    """
    Processes dependency in a POM file.

    Args:
        cache_data (Optional[dict]): Cache dictionary for storing dependency data, or None if disabled.
        config (Config): Configuration dictionary parsed from YAML.
        arguments (Arguments): Command-line arguments.
        dependency (ET.Element): Dependency.
        ns_mapping (dict): XML namespace mapping.
        root (ET.Element): Root element of the POM file.
        verify_ssl (bool): SSL verification flag.
        cve_data (dict[str, list[Vulnerability]]): CVE Data.
    """
    group, artifact = _utils.get_dependency_identifiers(dependency, ns_mapping)
    if not artifact or not group:
        logging.error("Missing artifactId or groupId in a dependency.")
        return

    version, skip_flag = _utils.get_version(config, arguments, ns_mapping, root, dependency)
    if skip_flag is True:
        _logutils.log_skip_if_required(config, arguments, group, artifact, version)
        return

    _logutils.log_search_if_required(config, arguments, group, artifact, version)

    _cveutils.log_vulnerability(config, arguments, group, artifact, version, cve_data)

    if cache_data is not None and cache_data.get(f"{group}:{artifact}") is not None:
        if _cache.process_cache_artifact(config, arguments, cache_data, artifact, group, version):
            return

    if not process_repositories(artifact, cache_data, config, group, arguments, verify_ssl, version):
        logging.warning(f"Not Found: {group}:{artifact}:{version}")


def process_repositories(
        artifact: str, cache_data: Optional[dict], config: Config, group: str,
        arguments: Arguments, verify_ssl: bool, version: Optional[str]
):
    """
    Processes repositories to find a dependency.

    Args:
        artifact (str): Artifact ID.
        cache_data (Optional[dict]): Cache data.
        config (Config): Parsed YAML as dict.
        group (str): Group ID.
        arguments (Arguments): Command-line arguments.
        verify_ssl (bool): SSL verification flag.
        version (Optional[str]): Dependency version.

    Returns:
        bool: True if the dependency is found, False otherwise.
    """
    if len(items := _config.config_items(config, 'repositories')):
        for section_key, repository_section in items:
            if (process_repository(
                    cache_data, config, arguments, group, artifact, version,
                    section_key, repository_section, verify_ssl)):
                return True
    return False


def process_modules_if_required(
        cache_data: Optional[dict], config: Config, arguments: Arguments, root: ET.Element,
        pom_path: str, ns_mapping: dict, prefix: Optional[str] = None
) -> None:
    """
    Processes modules in a POM file if required.

    Args:
        cache_data (Optional[dict]): Cache data.
        config (Config): Parsed YAML as dict.
        arguments (Arguments): Command-line arguments.
        root (ET.Element): Root element of the POM file.
        pom_path (str): Path to the POM file.
        ns_mapping (dict): XML namespace mapping.
        prefix (str, optional): Prefix for the artifact name.
    """
    if _config.get_config_value(config, arguments, 'process_modules'):
        directory_path = os.path.dirname(pom_path)
        modules = root.findall('.//xmlns:modules/xmlns:module', namespaces=ns_mapping)
        module_paths = [f"{directory_path}/{module.text}/pom.xml" for module in modules]
        valid_module_paths = [p for p in module_paths if p.startswith('http') or os.path.exists(p)]

        if _config.get_config_value(config, arguments, 'threading'):
            max_threads = _config.get_config_value(config, arguments, 'max_threads')
            with ThreadPoolExecutor(max_workers=max_threads) as executor:
                for future in as_completed([
                    executor.submit(process_pom, cache_data, config, arguments, module_path, prefix)
                    for module_path in valid_module_paths
                ]):
                    try:
                        future.result()
                    except Exception as e:  # pragma: no cover
                        logging.error(f"Error processing module: {e}")
        else:
            for module_path in valid_module_paths:
                process_pom(cache_data, config, arguments, module_path, prefix)


def process_artifact(
        cache_data: Optional[dict], config: Config, arguments: Arguments,
        artifact_to_find: str
) -> None:
    """
    Processes the search for a specified artifact.

    Args:
        cache_data (Optional[dict]): Cache data.
        config (Config): Parsed YAML as dict.
        arguments (Arguments): Command-line arguments.
        artifact_to_find (str): Artifact to search for in groupId:artifactId:version format.
    """
    verify_ssl = _config.get_config_value(config, arguments, 'verify', 'requests')
    group, artifact, version = artifact_to_find.split(':', maxsplit=2)

    _logutils.log_search_if_required(config, arguments, group, artifact, version)

    dependency_found = False
    for section_key, repository_section in _config.config_items(config, 'repositories'):
        if (dependency_found := process_repository(
                cache_data, config, arguments, group, artifact, version,
                section_key, repository_section, verify_ssl)):
            break
    if not dependency_found:
        logging.warning(f"Not Found: {group}:{artifact}, current:{version}")


def process_repository(
        cache_data: Optional[dict], config: Config, arguments: Arguments, group: str, artifact: str,
        version: Optional[str], section_key: str, repository_section: str, verify_ssl: bool
) -> bool:
    """
    Processes a repository section.

    Args:
        cache_data (Optional[dict]): Cache data.
        config (Config): Parsed YAML as dict.
        arguments (Arguments): Command-line arguments.
        group (str): Group ID.
        artifact (str): Artifact ID.
        version (Optional[str]): Artifact version.
        section_key (str): Repository section key.
        repository_section (str): Repository section name.
        verify_ssl (bool): SSL verification flag.

    Returns:
        bool: True if the dependency is found, False otherwise.
    """
    auth_info: Optional[tuple[str, str]] = None
    if _config.get_config_value(config, arguments, 'auth', repository_section, default=False):
        auth_info = (
            _config.get_config_value(config, arguments, 'user'),
            _config.get_config_value(config, arguments, 'password')
        )

    base_url = _config.get_config_value(config, arguments, 'base', repository_section)
    path_suffix = _config.get_config_value(config, arguments, 'path', repository_section)
    repository_name = _config.get_config_value(config, arguments, 'repo', repository_section)

    path = f"{base_url}/{path_suffix}"
    if repository_name is not None:
        path = f"{path}/{repository_name}"
    path = f"{path}/{group.replace('.', '/')}/{artifact}"

    with requests.Session() as session:
        metadata_url = path + '/maven-metadata.xml'
        response = session.get(metadata_url, auth=auth_info, verify=verify_ssl)

        if response.status_code == 200:
            tree = ET.ElementTree(ET.fromstring(response.text))
            version_elements = tree.getroot().findall('.//version')
            available_versions = [v.text for v in version_elements if v.text]
            available_versions.reverse()

            if _utils.check_versions(
                    cache_data, config, arguments, group, artifact, version, section_key,
                    path, auth_info, verify_ssl, available_versions, response):
                return True

    if _config.get_config_value(config, arguments, 'use_rest', repository_section, default=False):
        return process_rest(
            cache_data, config, arguments, group, artifact, version, section_key,
            repository_section, base_url, auth_info, verify_ssl)

    return False


def process_rest(
        cache_data: Optional[dict], config: Config, arguments: Arguments, group: str, artifact: str,
        version: Optional[str], section_key: str, repository_section: str, base_url: str,
        auth_info: Optional[tuple[str, str]], verify_ssl: bool
) -> bool:
    """
    Processes REST services for a repository.

    Args:
        cache_data (Optional[dict]): Cache data.
        config (Config): Parsed YAML as dict.
        arguments (Arguments): Command-line arguments.
        group (str): Group ID.
        artifact (str): Artifact ID.
        version (Optional[str]): Artifact version.
        section_key (str): Repository section key.
        repository_section (str): Repository section name.
        base_url (str): Base URL of the repository.
        auth_info (Optional[tuple[str, str]]): Authentication credentials.
        verify_ssl (bool): SSL verification flag.

    Returns:
        bool: True if the dependency is found, False otherwise.
    """
    repo = _config.get_config_value(config, arguments, 'repo', repository_section)
    path = f"{base_url}/service/rest/repository/browse/{repo}"
    path = f"{path}/{group.replace('.', '/')}/{artifact}"

    with requests.Session() as session:
        metadata_url = path + '/maven-metadata.xml'
        response = session.get(metadata_url, auth=auth_info, verify=verify_ssl)

        if response.status_code == 200:
            tree = ET.ElementTree(ET.fromstring(response.text))
            version_elements = tree.getroot().findall('.//version')
            available_versions = [v.text for v in version_elements if v.text]
            available_versions.reverse()

            if _utils.check_versions(
                    cache_data, config, arguments, group, artifact, version,
                    section_key, path, auth_info, verify_ssl, available_versions, response):
                return True

        response = session.get(path + '/', auth=auth_info, verify=verify_ssl)

        if response.status_code == 200:
            table = BeautifulSoup(response.text, 'html.parser').find('table')
            if table is None:  # pragma: no cover
                logging.error(f"Failed to parse versions from HTML at {path}")
                return False

            version_links = table.find_all('a')[1:]  # type: ignore
            path = f"{base_url}/repository/{repo}/{group.replace('.', '/')}/{artifact}"
            available_versions = [v.text for v in version_links if v.text]
            available_versions.reverse()

            if _utils.check_versions(
                    cache_data, config, arguments, group, artifact, version,
                    section_key, path, auth_info, verify_ssl, available_versions, response):
                return True

    return False



#!/usr/bin/python3
"""This file provides cve functions"""

import logging
import re
# noinspection PyPep8Naming
import xml.etree.ElementTree as ET
from dataclasses import dataclass
from itertools import islice
from typing import Optional

import maven_check_versions.cache as _cache
import maven_check_versions.config as _config
import maven_check_versions.utils as _utils
import requests
from maven_check_versions.config import Config, Arguments
from requests.auth import HTTPBasicAuth


@dataclass
class Vulnerability:
    """
    Vulnerability.
    """
    id: str  # NOSONAR # noqa: A003,VNE003
    displayName: Optional[str] = None  # NOSONAR # noqa: N815
    title: Optional[str] = None  # NOSONAR
    description: Optional[str] = None  # NOSONAR
    cvssScore: Optional[float] = None  # NOSONAR # noqa: N815
    cvssVector: Optional[str] = None  # NOSONAR # noqa: N815
    cve: Optional[str] = None  # NOSONAR
    cwe: Optional[str] = None  # NOSONAR
    reference: Optional[str] = None  # NOSONAR
    externalReferences: Optional[list] = None  # NOSONAR # noqa: N815
    versionRanges: Optional[list] = None  # NOSONAR # noqa: N815


def get_cve_data(
        config: Config, arguments: Arguments, dependencies: list[ET.Element],
        root: ET.Element, ns_mapping: dict
) -> dict[str, list[Vulnerability]]:
    """
    Retrieves CVE (Common Vulnerabilities and Exposures) data for the given dependencies
    using the OSS Index API, with caching support if enabled.

    Args:
        config (Config): Parsed YAML as dict.
        arguments (Arguments): Command-line arguments.
        dependencies (list[ET.Element]): Dependencies.
        root (ET.Element): Root element of the POM file.
        ns_mapping (dict): XML namespace mapping.

    Returns:
        dict[str, list[Vulnerability]]: CVE Data.
    """
    if not _config.get_config_value(config, arguments, 'oss_index_enabled', 'vulnerability', default=False):
        return {}

    coordinates = _get_coordinates(config, arguments, dependencies, ns_mapping, root)
    cve_data = _cache.load_cache(config, arguments, 'vulnerability') or {}

    for key, data in cve_data.items():
        cve_data[key] = [Vulnerability(**item) for item in data]

    coordinates = [coord for coord in coordinates if coord not in cve_data]

    if new_cve_data := _fetch_cve_data(config, arguments, coordinates):
        cve_data.update(new_cve_data)
        _cache.save_cache(config, arguments, cve_data, 'vulnerability')

    return cve_data


def log_vulnerability(
        config: Config, arguments: Arguments, group: str, artifact: str, version: Optional[str],
        cve_data: Optional[dict[str, list[Vulnerability]]]
) -> None:
    """
    Log vulnerability.

    Args:
        config (Config): Configuration dictionary parsed from YAML.
        arguments (Arguments): Command-line arguments.
        group (str): Group ID.
        artifact (str): Artifact ID.
        version (Optional[str]): Dependency version.
        cve_data (dict[str, Optional[list[Vulnerability]]]): CVE Data.
    """
    fail_score = _config.get_config_value(config, arguments, 'fail_score', 'vulnerability', default=0)
    cve_ref = _config.get_config_value(config, arguments, 'cve_reference', 'vulnerability', default=False)

    if cve_data is not None and (cves := cve_data.get(f"pkg:maven/{group}/{artifact}@{version}")):
        for cve in cves:
            info = f"cvssScore={cve.cvssScore} cve={cve.cve} cwe={cve.cwe} {cve.title}"
            if cve_ref:
                info = f"{info} {cve.reference}"
            logging.warning(f"Vulnerability for {group}:{artifact}:{version}: {info}")

            if cve.cvssScore and fail_score and cve.cvssScore >= fail_score:
                logging.error(f"cvssScore={cve.cvssScore} >= fail_score={fail_score}")
                raise AssertionError


def _get_coordinates(config, arguments, dependencies, ns_mapping, root) -> list[str]:
    """
    Get Coordinates.

    Args:
        config (Config): Parsed YAML as dict.
        arguments (Arguments): Command-line arguments.
        dependencies (list[ET.Element]): List of dependency elements from the POM file.
        root (ET.Element): The root element of the POM file's XML tree.
        ns_mapping (dict): A dictionary mapping XML namespaces for parsing.

    Returns:
        list[str]: Coordinates.
    """
    skip_no_versions = _config.get_config_value(
        config, arguments, 'skip_no_versions', 'vulnerability', default=False)
    combined = None
    if skip := _config.get_config_value(config, arguments, 'skip_checks', 'vulnerability'):
        combined = '(' + ')|('.join(skip) + ')'

    result: list = []
    for dependency in dependencies:
        (group, artifact) = _utils.get_dependency_identifiers(dependency, ns_mapping)
        (version, _) = _utils.get_version(config, arguments, ns_mapping, root, dependency)

        if skip_no_versions and version and re.match(r'^\${[^}]+}$', version):
            continue
        if combined is None or not re.match(combined, f"{group}:{artifact}:{version}"):
            result.append(f"pkg:maven/{group}/{artifact}@{version}")

    return result


def _oss_index_config(config: Config, arguments: Arguments) -> tuple:
    """
    Get OSS Index parameters.

    Args:
        config (Config): Parsed YAML as dict.
        arguments (Arguments): Command-line arguments.

    Returns:
        tuple: OSS Index parameters.
    """
    return (
        _config.get_config_value(
            config, arguments, 'oss_index_url', 'vulnerability',
            default='https://ossindex.sonatype.org/api/v3/component-report'),
        _config.get_config_value(config, arguments, 'oss_index_user', 'vulnerability'),
        _config.get_config_value(config, arguments, 'oss_index_token', 'vulnerability'),
        _config.get_config_value(
            config, arguments, 'oss_index_batch_size', 'vulnerability', default=128),
        _config.get_config_value(
            config, arguments, 'oss_index_keep_safe', 'vulnerability', default=False)
    )


def _fetch_cve_data(
        config: Config, arguments: Arguments, coordinates: list[str]
) -> dict[str, list[Vulnerability]]:
    """
    Get CVE data for coordinates.

    Args:
        config (Config): Parsed YAML as dict.
        arguments (Arguments): Command-line arguments.
        coordinates (list[str]): Coordinates.

    Returns:
        dict[str, list[Vulnerability]]: CVE Data.
    """
    result = {}
    try:
        url, user, token, batch_size, keep_safe = _oss_index_config(config, arguments)
        auth = HTTPBasicAuth(user, token)

        with requests.Session() as session:
            it = iter(coordinates)
            while batch := list(islice(it, batch_size)):
                response = session.post(url, json={"coordinates": batch}, auth=auth)
                if response.status_code != 200:
                    logging.error(f"OSS Index API error: {response.status_code}")
                    continue

                for item in response.json():
                    cves = []
                    if data := item.get('vulnerabilities'):
                        cves = [Vulnerability(**cve) for cve in data]
                    if len(cves) or keep_safe:
                        result.update({item['coordinates']: cves})

    except Exception as e:
        logging.error(f"Failed to _fetch_cve_data: {e}")
    return result
